"""Collision forces.

The main class in this file is Collision. It simulates Newtonian collisions
between two sprites. These collisions take into account the angles of the sprite
edges/vertices at the point of contact to realistically simulate the collision.
By default they respect Newtonian mechanics for rotation as well, taking into
consideration the moments of inertia of the sprites. However, there is an option
to ignore this and prohibit the collision from changing the angular velocity of
the sprites.

WARNING: Sprites must be star-shaped with respect to their center of mass (i.e.
the line fron the center of mass to any point in the sprite never exits the
sprite) for guaranteed correct collisions. If they are not star-shaped, there
can be rare events where two sprites collide at a corner and the collision is
not simulated correctly.
"""

# TODO(nwatters): Make collisions more robust when multiple collisions happen to
# a sprite in a single timestep. Consider using a backend like PyBullet or
# PyMunk.

# TODO(nwatters): Consider giving special treatment to circles for the collision
# point detection functions. Circular sprites are commonly used and have many
# vertices, which makes the _get_collision_vectors() function slow. Inferring
# collision point based on radius for circles might be worth making the code a
# little messier.

# TODO(nwatters): Handle the case where one sprite is dropped inside of another
# without having moved there by its physics. This can occur when the sprite is
# subject to some dynamics (e.g. via an action space or a portal) that directly
# change its position (instead of just changing its velocity). In this case the
# sprite should pop out of the overlap region. The current code does not handle
# this, but it is an important feature and would entail changing the
# _directed_collision_vectors() function.

from . import abstract_force
from matplotlib import path as mpl_path
from matplotlib import transforms as mpl_transforms
import numpy as np
from moog import sprite as sprite_lib

# Small float to ensure sprite position correction is conservative. Make sure
# this is not too small though, or ._make_disjoint() will be called
# unnecessarily often in Collision.step(). Something around 1e-2 seems to be
# fine.
_EPSILON = 1e-2


def _cross_2d(x, y):
    """2-dimensional cross product.
    
    Args:
        x: Numpy array of shape [..., 2]. First vector.
        y: Numpy array of shape [..., 2]. Second vector.
        
    Returns:
        cross: Numpy array of shape [...]. Cross product of x and y.
    """
    return x[..., 0] * y[..., 1] - x[..., 1] * y[..., 0]


def _relative_motion_trajectory(path, path_sprite, anchor_sprite, delta_t):
    """Find trajectory of a path in the coordinate frame of an anchor sprite.

    Here path is rigidly tethered to path_sprite, i.e. it may consist of
    vertices of path_sprite. This function produces the trajectory of the path
    points in the inertial frame of anchor_sprite, i.e. the moving coordinate
    system in which anchor_sprite is static.

    Args:
        path: Instance of mpl_path.Path. path.vertices has shape [N, 2].
        path_sprite: Instance of ../sprite.Sprite.
        anchor_sprite: Instance of ../sprite.Sprite.
        delta_t: Scalar. Timestep over which to produce the trajectories.

    Returns:
        trajectory: Numpy array of shape [N, 2, 2]. Element i of trajectory
            is an array of shape [2, 2] in which the second element is the i'th
            path vertex and the first element is the previous position of that
            vertex in the relative coordinate transformation between
            anchor_sprite and path_sprite.
    """
    transform = (
        mpl_transforms.Affine2D().rotate_around(
            *path_sprite.position,
            theta=-1 * path_sprite.angle_vel * delta_t) +
        mpl_transforms.Affine2D().translate(
            *(-1 * path_sprite.velocity * delta_t)) +
        mpl_transforms.Affine2D().rotate_around(
            *anchor_sprite.position,
            theta=anchor_sprite.angle_vel * delta_t) +
        mpl_transforms.Affine2D().translate(
            *anchor_sprite.velocity * delta_t)
    )
    previous_path = transform.transform_path(path)
    trajectory = np.stack((previous_path.vertices, path.vertices), axis=1)

    return trajectory


def _directed_collision_vectors(sprite_0, sprite_1, delta_t):
    """Find single contact point and collision vectors.
    
    This function does the following:
        1. Finds all vertices of sprite_0 that lie inside of sprite_1.
        2. Finds the true continuous-time point of contact for each of these.
        3. Selects on that is most likely to be the true collisions point.
        4. Produces that collision point, the collision normal vector, and
            relative displacement of the sprites since the true collision time.

    Note that this is directed, i.e. only considers vertices of sprite_0 that
    lie in sprite_1. In reality a collision might occur by a vertex of sprite_1
    initially entering sprite_0. Hence we must call this function on
    *(sprite_1, sprite_0) as well. See _get_collision_vectors() for details.

    Args:
        sprite_0: Instance of ../sprite.Sprite.
        sprite_1: Instance of ../sprite.Sprite.
        delta_t: Scalar. Timestep of the physics.

    Returns:
        collision_point: None or numpy array of shape [2]. Position of the
            collision.
        collision_normal: None or numpy array of shape [2]. Zero-centered vector
            indicating the normal of the collision surface, i.e. the direction
            in which the collision force is applied. It is signed, so we let it
            point toward the sprite_1 side of the collision.
        since_collision: None or numpy array of shape [2]. Relative displacement
            of the two sprites since the collision. It is signed, so we let it
            point toward the sprite_0 side of the collision, i.e. representing
            how much sprite_0 has moved relative to sprite_0 since the collision
            event.
        perpendicular: A scaling of collision_normal, where the length is the
            margin of overlap of the sprites. This can be used to make the
            sprites disjoint.
    """
    # First find points of sprite_0 inside sprite_1
    vertices_0 = sprite_0.vertices
    contained_inds = np.argwhere(sprite_1.contains_points(vertices_0))[:, 0]
    if len(contained_inds) == 0:
        return None, None, None, None
    contained_vertices_0 = vertices_0[contained_inds]

    # Now find previous-to-current timestep trajectory of each contained vertex
    traj = _relative_motion_trajectory(
        mpl_path.Path(contained_vertices_0),
        sprite_0,
        sprite_1,
        delta_t,
    )

    # We will now find the crossing points between the previous-to-current
    # timestep relative trajectories of the vertices of sprite 0 and the edges 
    # of sprite 1. Instead of restricting ourselves to only the crossings of
    # those segments, we also extrapolate the relative trajectory of sprite 0
    # backwards in time to consider crossing points which may have occured
    # multiple timesteps ago. This is important to do because the relative
    # velocity of the sprites may have changed recently (e.g. due to some force
    # or action), so considering multi-timestep historical crossings avoids
    # instabilities. We can easily compute these historical crossings from the
    # segment crossing coefficients.
    vertices_1 = sprite_1.vertices
    vertices_1_next = np.concatenate((vertices_1[1:], vertices_1[:1]), axis=0)
    cross_a, cross_b = sprite_lib.segment_crossing_coefficients(
        start_0=traj[:, 0],
        end_0=traj[:, 1],
        start_1=vertices_1,
        end_1=vertices_1_next,
    )

    # To find current segment crossings we would further require cross_a > 0 and
    # cross_a < 1, but by allowing cross_a to be negative we let in historical
    # crossings and by allowing cross_a to be greater than 1 we let in future
    # crossings.
    crossings = (cross_b >= 0) * (cross_b <= 1)

    if not np.any(crossings):
        # This can happen because of _EPSILON_INTERPOLATION in sprite.py
        return None, None, None, None
    
    # We're going to use cross_a values to find the most eggregious crossing, so
    # set the non-crossing entries to -Inf. This is less cumbersome to implement
    # than setting to NaN and use np.nanargmin, because some slices in np.argmax
    # below might be all -Inf.
    cross_a[np.logical_not(crossings)] = -np.inf
    abs_cross_a = np.abs(1. - cross_a)

    # Get the crossing closest to the current vertex position for each vertex in
    # sprite_0. This will either be the most recent crossing or the most
    # imminent crossing.
    inds_crossings = np.argmin(abs_cross_a, axis=1)
    crossing_points = np.array([
        traj[i_0, 0] + cross_a[i_0, i_1] * (traj[i_0, 1] - traj[i_0, 0])
        for (i_0, i_1) in enumerate(inds_crossings)
    ])

    # Get segment from crossing point to trajectory end
    diffs_from_crossings = np.array([
        traj[i, 1] - crossing_points[i]
        for i in range(len(crossing_points))
    ])

    # Now we try to deduce which of the crossing points is the true collision
    # point. We do this by selecting the one for which the trajectory segment
    # endpoint is furthest from the crossing point, i.e. the point that is
    # deepest inside sprite_1.
    dists_vertices_crossings = np.linalg.norm(diffs_from_crossings, axis=1)
    dists_vertices_crossings[dists_vertices_crossings == np.inf] = 0
    crossing_points_ind = np.argmax(dists_vertices_crossings)
    sprite_1_ind = inds_crossings[crossing_points_ind]
    collision_point = crossing_points[crossing_points_ind]
    since_collision = diffs_from_crossings[crossing_points_ind]

    # If the true collision point is in the future, don't collide.
    if cross_a[crossing_points_ind, sprite_1_ind] > 1:
        # True collision point is in the future
        return collision_point, np.nan, since_collision, None

    # Finally, compute the collision vector normal to sprite_1
    delta_vertex_1 = (vertices_1_next[sprite_1_ind] -
                      vertices_1[sprite_1_ind])
    normal_vector = np.array([delta_vertex_1[1], -1 * delta_vertex_1[0]])
    collision_normal = normal_vector / np.linalg.norm(normal_vector)

    # Compute the perpendicular margin of overlap
    projection = delta_vertex_1 * (
        np.dot(since_collision, delta_vertex_1) /
        np.dot(delta_vertex_1, delta_vertex_1)
    )
    perpendicular = since_collision - projection

    return collision_point, collision_normal, since_collision, perpendicular


def _get_collision_vectors(sprite_0, sprite_1, delta_t):
    """Get contact point, collision normal, and displacement since collision.
    
    This function produces the collision point, collision normal, and relative
    displacement since the collision event for two sprites. It approximately
    infers these as  if the sprites had continuous time, i.e. the collision
    point is the true point of collision between the previous and current
    timestep.

    The collision point, normal, and displacement (plus things like the sprite
    centers of mass and moments of inertia) of are all that are needed from the
    sprite geometry for computing the physics. Once these are computed, we
    never have to look at the sprite vertices again.

    Args:
        sprite_0: Instance of ../sprite.Sprite.
        sprite_1: Instance of ../sprite.Sprite.
        delta_t: Scalar. Timestep of the physics.

    Returns:
        collision_point: None or numpy array of shape [2]. Position of the
            collision.
        collision_normal: None or numpy array of shape [2]. Zero-centered vector
            indicating the normal of the collision surface, i.e. the direction
            in which the collision force is applied. It is signed, so we let it
            point toward the sprite_0 side of the collision.
        since_collision: None or numpy array of shape [2]. Relative displacement
            of the two sprites since the collision. It is signed, so we let it
            point toward the sprite_1 side of the collision, i.e. representing
            how much sprite_0 has moved relative to sprite_1 since the collision
            event.
    """
    # First find the collision point on sprite_0 boundary
    collision_point_0, collision_normal_0, since_collision_0, perp_0 = (
        _directed_collision_vectors(sprite_1, sprite_0, delta_t))
    
    # Now find the collision point on sprite_1 boundary
    collision_point_1, collision_normal_1, since_collision_1, perp_1 = (
        _directed_collision_vectors(sprite_0, sprite_1, delta_t))

    # Make since_collision_i zero if no collision happened in the i direction
    if collision_point_0 is None:
        since_collision_0 = np.zeros(2)
    else:
        # Must negate collision_normal and since_collision for symmetry
        collision_normal_0 = -1. * collision_normal_0
        since_collision_0 = -1. * since_collision_0
    if since_collision_1 is None:
        since_collision_1 = np.zeros(2)

    # Pick which collision point to treat as the real collision point
    if np.linalg.norm(since_collision_0) > np.linalg.norm(since_collision_1):
        return collision_point_0, collision_normal_0, since_collision_0, perp_0
    else:
        return collision_point_1, collision_normal_1, since_collision_1, perp_1


def _collide_without_update_angle_vel(sprite_0,
                                      sprite_1,
                                      collision_point,
                                      collision_normal,
                                      elasticity=1.,
                                      symmetric=True):
    """Simulate collision in which sprite angular velocity cannot change.

    Since the sprite angular velocity is prohibited from changing, this
    collision is non-Newtonion, i.e. does not conserve angular momentum. However
    it does converse kinetic enery and momentum, hence models what the collision
    would be if the collision point were between the two sprites' centers of
    mass.

    Args:
        sprite_0: Instance of ../sprite.Sprite. If symmetric = False, this
            sprite is the one that is doing the colliding, i.e. this one's
            velocity changes.
        sprite_1: Instance of ../sprite.Sprite. If symmetric = False, this
            sprite is treated as having infinite mass (i.e. has fixed velocity).
        collision_point: Numpy array of shape [2]. Position of the collision.
        collision_normal: Numpy array of shape [2]. Unit normal of the
            collision, pointing towards sprote_0. Must have norm 1.
        elasticity: Float in [0, 1]. Elasticity 1 means the sprites bounce
            fully. Elasticity 0 means they stick together completely.
        symmetric: Bool. If True, collision is symmetric and both sprites'
            velocities are updated. If False, sprite_1 is treated as having
            infinite mass.
    """
    # Sanity check that collision_normal has norm 1
    collision_normal_norm = np.linalg.norm(collision_normal)
    if not np.isclose(collision_normal_norm, 1., atol=1e-4):
        raise ValueError(
            'collision_normal_norm is {}, which is not close to 1.'.format(
                collision_normal_norm))

    vel_0 = sprite_0.velocity
    vel_1 = sprite_1.velocity
    m_0 = sprite_0.mass
    m_1 = sprite_1.mass

    # Only consider the component of velocity parallel to normal
    vel_0_normal = np.dot(vel_0, collision_normal) * collision_normal
    vel_1_normal = np.dot(vel_1, collision_normal) * collision_normal

    # Find inertial reference frame, i.e. velocity of center of mass of the
    # entire system
    if symmetric:
        vel_cm_normal = (vel_0_normal * m_0 + vel_1_normal * m_1) / (m_0 + m_1)
    else:
        vel_cm_normal = vel_1_normal

    # Reflect each velocity across vel_cm, with elasticity
    delta_v_0 = (1 + elasticity) * (vel_cm_normal - vel_0_normal)
    delta_v_1 = (1 + elasticity) * (vel_cm_normal - vel_1_normal)

    # Update the sprite velocities
    sprite_0.velocity += delta_v_0
    sprite_1.velocity += delta_v_1


def _collide_with_update_angle_vel(sprite_0,
                                   sprite_1,
                                   collision_point,
                                   collision_normal,
                                   elasticity=1.,
                                   symmetric=True):
    """Simulate collision.

    This simulates Newtonian collisions between sprite_0 and sprite_1. It takes
    into account the geometry of the sprites, collision contact point, collision
    normal at the contact point, rotational inertia of the sprites, etc. Hence
    the collisions this simulates are very realistic.

    Deriving the equations governing the collision dynamics is non-trivial, so
    we sketch the approach here:
        
        First, for any collision between a rigid body with moment of unertia I,
        angular velocity w, mass M, and center-of-mass velocity v, we have
            I delta(w) = M delta(v) r sin(theta)
        where r is the distance between the collision point and the center of
        mass and theta is the angle between the collision normal and the center
        of mass. This equation is independent of the object with which the
        collision happens, and can be derived from conservation of momentum and
        conservation of angular momentum given some point mass being collided
        with.

        Now suppose we have two rigid bodies with subscripts 0 and 1. Then the
        above equation gives us
            I_0 delta(w_0) = M_0 delta(v_0) r_0 sin(theta_0)
            I_1 delta(w_1) = M_1 delta(v_1) r_1 sin(theta_1)
        Now conversation of global momentum gives us
            0 = sum_i[M_i delta(v_i)]
        And conservation of global kinetic energy gives
            0 = sum_i[M_i (v_i_after^2 - v_i_before^2) +
                      I_i (w_i_after^2 - w_i_before^2)]
              = sum_i[M_i delta(v_i) gamma(v_i) + I_i delta(w_i) gamma(w_i)]
        where gamma(x_i) = x_i_after + x_i_before.
        
        We must now solve these four equations for the four variable
            {v_i_after, w_i_after}_{i = 0, 1}
        Substituting gamma(x_i) = delta(x_i) + 2 x_i_before into the kinetic
        energy equation and eliminating all appearances of delta(v_1) and
        delta(w_1) using the previous equations, we end up with the solution in
        the code.

    Args:
        sprite_0: Instance of ../sprite.Sprite. If symmetric = False, this
            sprite is the one that is doing the colliding, i.e. this one's
            velocity changes.
        sprite_1: Instance of ../sprite.Sprite. If symmetric = False, this
            sprite is treated as having infinite mass and infinite moment of
            inertia (i.e. has fixed velocity and angular velocity).
        collision_point: Numpy array of shape [2]. Position of the collision.
        collision_normal: Numpy array of shape [2]. Unit normal of the
            collision, pointing towards sprote_0. Must have norm 1.
        elasticity: Float in [0, 1]. Elasticity 1 means the sprites bounce
            fully. Elasticity 0 means they stick together completely.
        symmetric: Bool. If True, collision is symmetric and both sprites'
            velocities are updated. If False, sprite_1 is treated as having
            infinite mass and infinite moment of inertia.
    """

    # Alias some scalars that we'll need
    m_0 = sprite_0.mass
    m_1 = sprite_1.mass
    w_0 = sprite_0.angle_vel
    w_1 = sprite_1.angle_vel
    i_0 = sprite_0.moment_of_inertia
    i_1 = sprite_1.moment_of_inertia

    # Extract only magnitude of velocity component parallel to normal
    v_0 = np.dot(sprite_0.velocity, collision_normal)
    v_1 = np.dot(sprite_1.velocity, collision_normal)

    # Extract sin(theta), where theta is angle between normal and contact point
    c_point_0 = collision_point - sprite_0.position
    c_point_1 = collision_point - sprite_1.position
    r_0 = np.linalg.norm(c_point_0)
    r_1 = np.linalg.norm(c_point_1)
    sin_theta_0 = _cross_2d(c_point_0, collision_normal) / r_0
    sin_theta_1 = _cross_2d(c_point_1, collision_normal) / r_1

    # Apply the collision equations. See docstring for a derivation sketch.
    s_0 = r_0 * sin_theta_0
    s_1 = r_1 * sin_theta_1

    a = m_0 + m_1 + m_0 * m_1 * ((s_0 * s_0 / i_0) + (s_1 * s_1 / i_1))
    b = (1 + elasticity) * (v_0 - v_1 + w_0 * s_0 - w_1 * s_1)
    if symmetric:
        delta_v_0 = -1 * m_1 * b / a
        delta_v_1 = m_0 * b / a
    else:
        delta_v_0 = -1 * m_1 * b / (a - m_0)
        delta_v_1 = 0.
    delta_w_0 = m_0 * delta_v_0 * s_0 / i_0
    delta_w_1 = m_1 * delta_v_1 * s_1 / i_1

    # Update sprite velocities and angular velocities
    sprite_0.velocity += delta_v_0 * collision_normal
    sprite_1.velocity += delta_v_1 * collision_normal
    sprite_0.angle_vel += delta_w_0
    sprite_1.angle_vel += delta_w_1


class Collision(abstract_force.AbstractForce):
    """Collision simulator.
    
    This class simulates Newtonian collisions between two rigid bodies, namely
    sprites in the environment. The physics assumes that the sprites have
    uniform density, as this is important for calculating their moments of
    inertia (see ../sprite for details).
    """

    def __init__(self, elasticity=1., symmetric=False, update_angle_vel=True,
                 max_recursion_depth=0):
        """Constructor.
        
        Args:
            elasticity: Float in [0, 1]. Elasticity 1 means the sprites bounce
                fully. Elasticity 0 means they stick together completely.
            symmetric: Bool. If True, collision is symmetric and both sprites
                are updated. If False, sprite_1 in .step() is treated as having
                infinite mass. This is useful for modeling bounces off of
                obstructors/walls that are fixed in the environment.
            update_angle_vel: Bool. If True, fully simulate the rotational
                mechanics and update the sprites angular velocity. If False,
                sprite angular velocities are not updated and the collision is
                treated as if the contact point was between the two sprites'
                centers of mass. Simulating the full rotational mechanics is
                slightly slower.
            max_recursion_depth: Int. Number of times self.step() may recurse.
                Usually, 0 (no recursion) is fine. However, larger values make
                the collisions slightly more accurate (at the expense of runtime
                efficiency) if multiple collisions are happening within a single
                step, e.g. colliding into a corner.
        """
        self._elasticity = elasticity
        self._symmetric = symmetric
        self._update_angle_vel = update_angle_vel
        self._max_recursion_depth = max_recursion_depth

    def step(self, sprite_0, sprite_1, updates_per_env_step, recursion_depth=0):
        """Step the physics.
        
        Args:
            sprite_0: Instance of ../sprite.Sprite. If self._symmetric = False,
                this sprite is the one that is doing the colliding, i.e. this
                one's velocity changes.
            sprite_1: Instance of ../sprite.Sprite. If self._symmetric = False,
                this sprite is treated as having infinite mass.
            updates_per_env_step: Int. Number of times this force step is called
                for each step of the physics in the environment. This is used
                here for inferring the exact contact point in the collision.
            recursion_depth: Int. Number of times this function has recursed.
                Used internally only to catch and avoid max recursion depth
                errors.
        """
        if recursion_depth > self._max_recursion_depth:
            return

        if sprite_0 == sprite_1:
            return

        if not sprite_0.overlaps_sprite(sprite_1):
            return

        delta_t = 1. / updates_per_env_step

        # First get collision point, normal, and relative sprite displacement
        # since the collision
        collision_point, collision_normal, _, perpendicular = (
            _get_collision_vectors(sprite_0, sprite_1, delta_t))
        
        if collision_point is None:
            # Although the sprites do overlap, neither sprite contains any of
            # the other's vertices. This can happen when the sprites collide
            # exactly at two corners. This is annoying to handle because we must
            # infer which corner hit which face in between timesteps, but must
            # be done to ensure stability. See self._make_disjoint() for
            # details.
            self._make_disjoint(sprite_0, sprite_1)
        else:
            # Whether the collision point is in the future, in which case we
            # leave the sprites alone.
            future_collision_point = (
                np.isscalar(collision_normal) and np.isnan(collision_normal))

            if not future_collision_point:
                # There was a collision in the past, so we must simulate it
                # First, displace the sprites so they are no longer intersecting
                # Note that instead of perpendicular displacement we could use
                # since_collision. However, sometimes since_collision can have a
                # very acute angle and large magnitude due to angular velocity
                # or multiple collisions, so in prectice displacing by the
                # perpendicular is more stable.
                if self._symmetric:
                    sprite_0.position = (
                        sprite_0.position - (0.5 + _EPSILON) * perpendicular)
                    sprite_1.position = (
                        sprite_1.position + (0.5 + _EPSILON) * perpendicular)
                else:
                    sprite_0.position = (
                        sprite_0.position - (1. + _EPSILON) * perpendicular)

                # Second, change sprite velocities and angular velocities per
                # Newtonian physics
                if self._update_angle_vel:
                    _collide_with_update_angle_vel(
                        sprite_0,
                        sprite_1,
                        collision_point,
                        collision_normal,
                        elasticity=self._elasticity,
                        symmetric=self._symmetric,
                    )
                else:
                    _collide_without_update_angle_vel(
                        sprite_0,
                        sprite_1,
                        collision_point,
                        collision_normal,
                        elasticity=self._elasticity,
                        symmetric=self._symmetric,
                    )
            else:
                return
            
        # After perturbing the sprite positions and velocities, we recursively
        # step again in case that perturbation has now created another
        # collision.
        self.step(sprite_0, sprite_1, updates_per_env_step,
                  recursion_depth=recursion_depth + 1)

    def _make_disjoint(self, sprite_0, sprite_1):
        """Perturb the positions of sprite_0 and sprite_1 to make them disjoint.

        In theory, with infinitesimal delta_t, the collision detection would
        work without this correction since every collision occurs by a vertex of
        one sprite entering another sprite. However, with a non-zero delta_t, if
        a collision occurs almost exactly at two sharp corners of sprites, the
        vertex entrypoint could be missed due to the time discretization. So
        this correction will catch that case and perturb the sprites so they no
        longer intersect.

        Furthermore, sometimes if collisions with multiple sprites happen within
        one physics timestep, the collision misses one of them, so this
        perturbation comes in handy then to essentially push one of the
        collisions off until the next timestep.

        Finally, sometimes a sprite position may be macigally changed (e.g. by a
        discrete action space) to make it enter another sprite without moving
        there by its physics, so this correction is useful to separate the two
        overlapping sprites in that case as well.

        The way this function works is a bit ugly and not easy to describe. If
        you want to understand it, please draw out some examples on paper and
        see what the code does with them.

        Args:
            sprite_0: Instance of ../sprite.Sprite.
            sprite_1: Instance of ../sprite.Sprite.

        Returns:
            Boolean indicating whether the sprites were overlapping. This is
            useful for the calling code to know whether the sprie positions have
            been perturbed.
        """
        crossing_points, inds_crossings = sprite_lib.sprite_edge_crossings(
            sprite_0, sprite_1)

        if len(inds_crossings) <= 1:
            return True

        # correction_i is how much sprite_i must move (independent of sprite_j)
        # for the two to be disjoint.
        correction_0 = _position_correction(
            crossing_points,
            sprite_0,
            inds_crossings[:, 0],
            sprite_1,
            inds_crossings[:, 1])
        correction_1 = _position_correction(
            crossing_points,
            sprite_1,
            inds_crossings[:, 1],
            sprite_0,
            inds_crossings[:, 0])

        if np.linalg.norm(correction_0) > np.linalg.norm(correction_1):
            correction = -1 * (1 + _EPSILON) * correction_0
        else:
            correction = (1 + _EPSILON) * correction_0
        
        if not all(np.isfinite(correction)):  # no correction needed
            correction = np.zeros(2)
        
        if self._symmetric:
            sprite_0.position = sprite_0.position + 0.5 * correction
            sprite_1.position = sprite_1.position - 0.5 * correction
        else:
            sprite_0.position = sprite_0.position + correction
        
        return False


def _position_correction(crossing_points,
                         sprite_0,
                         crossing_inds_0,
                         sprite_1,
                         crossing_inds_1):
    """Finds a correction for the position to make the sprites disjoint.
    
    This finds how much the sprite_0 must move to alleviate the crossing points
    closest to its center of mass.

    Args:
        crossing_points: Numpy float array of size [K, 2].
        sprite_0: Instance of ../sprite.Sprite.
        crossing_inds_0: Numpy int array of size [K, 2].
        sprite_1: Instance of ../sprite.Sprite.
        crossing_inds_1: Numpy int array of size [K, 2].
    """
    vertices = sprite_0.vertices
    other_vertices = sprite_1.vertices

    # Sort crossing points by their proximity to the sprite_0's center of mass.
    dists_from_cm = np.linalg.norm(crossing_points - sprite_0.position, axis=1)
    sorted_inds = np.argsort(dists_from_cm)

    if sorted_inds[0] == sorted_inds[1]:
        # sprite_0 has no offending vertices, so sees no correction itself
        return np.array([np.inf, np.inf])

    # Get the two closest points. Ultimately we will find a perturbation
    # perpendicular to the segment between these two points. Specifically, we
    # will find the sprite_0 vertex closest to the other sprite_0's center of
    # mass along the direction perpendicular to the segment between these two
    # points. That vertex is the most eggregious offender and will tell use how
    # much we have to perturb.
    pt_0 = crossing_points[sorted_inds[0]]
    pt_1 = crossing_points[sorted_inds[1]]

    def _get_norm_v(point_0, point_1):
        # Get normalized vector perpendicular to [point_0, point_1] and pointing
        # towards sprite_1.position
        normalized_bdry = point_1 - point_0
        normalized_bdry /= np.linalg.norm(normalized_bdry)
        norm_v = normalized_bdry * -1 * np.sign(
            np.dot(sprite_1.position - point_0, normalized_bdry))
        return norm_v

    # Get the normal vector from the [pt_0, pt_1] segment in the direction of
    # cm_other.
    if crossing_inds_1[sorted_inds[0]] == crossing_inds_1[sorted_inds[0]]:
        # In this case using the edge of sprite_1 is more accurate
        pt_0_ind = crossing_inds_1[sorted_inds[0]]
        pt_1_ind = (pt_0_ind - 1) % len(other_vertices)
        norm_v = _get_norm_v(other_vertices[pt_1_ind], other_vertices[pt_0_ind])
    else:
        norm_v = _get_norm_v(pt_0, pt_1)

    # We're now going to traverse through some of sprite_0's vertices looking
    # for the one furthest in the norm_v direction. However, we don't want to
    # consider all of sprite_0's vertices (that could give wrong results for
    # some star-shaped sprites). We only want to consider those that lie between
    # the crossing points of interest. So we do this by finding which endpoint
    # of the first crossing point's sprite_0 edge lies in the norm_v direction
    # and traverse from there until we are in the negative norm_v direction.
    ind_forward = crossing_inds_0[sorted_inds[0]]
    ind_backward = (ind_forward - 1) % len(vertices)
    if np.dot(vertices[ind_forward] - pt_0, norm_v) > 0:
        # The forward endpoint of this sprite_0 edge is in the norm_v direction
        parity = 1
        current_ind = ind_forward
    elif np.dot(vertices[ind_backward] - pt_0, norm_v) > 0:
        # The backward endpoint of this sprite_0 edge is in the norm_v direction
        parity = -1
        current_ind = ind_backward
    else:
        # Neither vertex is in the norm_v direction, in which case we aren't
        # actually overlapping the other sprite_0 (probably just barely touching
        # it, perhaps caused by _EPSILON)
        return np.array([np.inf, np.inf])

    # Now iterate through vertices looking for the most eggregious offender
    # until we're no longer in the norm_v direction.
    worst_penalty = 0
    while np.dot(vertices[current_ind] - pt_0, norm_v) > 0:
        penalty = np.dot(vertices[current_ind] - pt_0, norm_v)
        if penalty > worst_penalty:
            worst_penalty = penalty
        current_ind = (current_ind + parity) % len(vertices)
    
    correction = worst_penalty * norm_v

    return correction



"""Rules that move sprites from one layer to another."""

from . import abstract_rule
import numpy as np


class ChangeLayer(abstract_rule.AbstractRule):
    """ChangeLayer rule.
    
    This rule is used to change the layer of a sprite based on a filter of the
    sprite. For example, it could be used to change a sprite's layer if the
    sprite has changed to a new color.
    """

    def __init__(self, old_layer, new_layer, filter_fn=None):
        """Constructor.

        Args:
            old_layer: String. Must be a key in the environment state. Sprites
                in this layer will be considered for the layer change.
            new_layer: String. Must be a key in the environment state. Layer to
                which sprites in old_layer may be moved.
            filter_fn: Function sprite -> bool. Whether to move the sprite from
                layer old_layer to layer new_layer. Defaults to always True.
        """
        self._old_layer = old_layer
        self._new_layer = new_layer
        
        if filter_fn is None:
            self._filter_fn =  lambda s: True
        else:
            self._filter_fn = filter_fn

    def step(self, state, meta_state):
        """Apply rule, potential moving sprites from old to new layer."""
        del meta_state
        
        should_change = [self._filter_fn(s) for s in state[self._old_layer]]
        change_inds = np.argwhere(should_change)[:, 0]

        count_changed_already = 0
        for i in change_inds:
            state[self._new_layer].append(
                state[self._old_layer].pop(i - count_changed_already))
            count_changed_already += 1



"""Rule that is applied only when a condition is satisfied.

The main class here is ConditionalRule, which applies a rule or rules X number
of times, where X is a bool or an integer returned by a function applied to the
state.
"""

from . import abstract_rule
import inspect


class ConditionalRule(abstract_rule.AbstractRule):
    """Apply a rule X number of times, where X is a function of the state.

    X may be a boolean, indicating whether to apply the rule once or not.

    For example, here is code for a rule that creates a new prey every time a
    prey is contacted by an agent:
    ```python
        my_rule = ConditionalRule(
            condition=get_contact_counter('prey', 'agent'),
            rules=CreateSprites(
                layer='prey',
                generator=prey_generator,
                without_overlapping=['walls', 'agent']),
        )
    ```
    """

    def __init__(self, condition, rules):
        """Constructor.

        Args:
            condition: Function with one of the following signatures:
                    * state --> bool or int
                    * state, meta_state --> bool or int
                If output is bool, indicates whether to apply rule. If output is
                int, indicates how many times to apply rule.
            rules: Instance of abstract_rule.AbstractRule or iterable of such.
        """
        if len(inspect.signature(condition).parameters.values()) == 1:
            self._condition = lambda state, meta_state: condition(state)
        else:
            self._condition = condition
        
        if not isinstance(rules, (list, tuple)):
            self._rules = [rules]
        else:
            self._rules = rules

    def reset(self, state, meta_state):
        for rule in self._rules:
            rule.reset(state, meta_state)

    def step(self, state, meta_state):
        for _ in range(self._condition(state, meta_state)):
            for rule in self._rules:
                rule.step(state, meta_state)



"""Portal.

The entrypoint is Portal, a rule that teleports a sprite from one portal sprite
position to another.
"""

from . import abstract_rule
import numpy as np


class Portal(abstract_rule.AbstractRule):
    """Makes a sprite teleport if it enters a portal sprite."""

    def __init__(self, teleporting_layer, portal_layer):
        """Constructor.

        The environment state must have an even number of sprites in
        portal_layer, because portals are paired up in order. I.e. if there are
        4 portals, the first two will teleport to each other and the second two
        will teleport to each other.

        Also, once a sprite has teleported, it cannot immediately teleport again
        until it exits the portal sprite. This is kept track of by
        self._currently_teleporting, and is necessary to prevent a sprite from
        immediately teleporting back and forth between portals.

        Args:
            teleporting_layer: String. Must be a key in the environment state.
                Sprites in this layer will be teleported from the state if their
                position enters a sprite in portal_layer.
            portal_layer: String. Must be a key in the environment state.
        """        
        self._teleporting_layer = teleporting_layer
        self._portal_layer = portal_layer

    def reset(self, state, meta_state):
        del state
        del meta_state
        self._currently_teleporting = set()

    def step(self, state, meta_state):
        """Apply rule to state.

        Teleport sprites in teleporting layer if they have entered a portal
        sprite, unless they are in self._currently_teleporting.
        """
        del meta_state

        portals = state[self._portal_layer]
        num_portals = len(portals)
        if num_portals % 2 != 0:
            raise ValueError(
                'There must be an even number of portals, but you have {} '
                'portals.'.format(num_portals))
        
        for sprite in state[self._teleporting_layer]:
            in_portals = [portal.contains_point(sprite.position)
                          for portal in portals]
            in_portal_inds = np.argwhere(in_portals)[:, 0]

            if len(in_portal_inds) == 0:
                # Sprite is not in any portal, so make sure we don't think
                # sprite is currently teleporting
                self._currently_teleporting.discard(sprite.id)
                continue

            if sprite.id in self._currently_teleporting:
                # To prevent immediately teleporting back and forth between
                # portals
                continue
            
            # Teleport the sprite
            entry_ind = in_portal_inds[0]
            exit_ind = entry_ind - 1 if entry_ind % 2 else entry_ind + 1
            sprite.position = np.copy(portals[exit_ind].position)
            self._currently_teleporting.add(sprite.id)



"""Maze task with predators, prey, and boosters.

The predators (red circles) chase the agent. The agent receives reward for
catching prey (yellow circles), which disappear upon capture. The boosters (blue
triangles) temporarily increase the agent's speed. The portals (white squares) 
teleport the agent from one place to another.
"""

import collections
import numpy as np

from moog import action_spaces
from moog import game_rules
from moog import observers
from moog import physics as physics_lib
from moog import shapes
from moog import sprite
from moog import tasks
from moog.state_initialization import distributions as distribs
from moog.state_initialization import sprite_generators


class Booster(game_rules.AbstractRule):
    """Game rule for boosters.

    This rule makes the agent temporarily brigher in color and lighter in mass
    (to move faster) upon contact with a booster sprite.
    """

    def __init__(self,
                 mass_multiplier=0.4,
                 c2_multiplier=0.1,
                 boost_duration=60,
                 agent_layer='agent',
                 booster_layer='boosters'):
        """Constructor.

        Args:
            mass_multiplier: Float. Mass of agent after contacting a booster.
            c2_multiplier: Float. Brightness increase factor upon contacting a
                booster.
            boost_duration: Int. Number of steps before boost ends.
            agent_layer: String. Name of the agent layer in the state.
            booster_layer: String. Name of the booster layer in the state.
        """
        self._mass_multiplier = mass_multiplier
        self._c2_multiplier = c2_multiplier
        self.boost_duration = boost_duration
        self._agent_layer = agent_layer
        self._booster_layer = booster_layer

    def reset(self, state, meta_state):
        del state
        del meta_state
        self._steps_until_expire = np.inf

    def _apply_change(self, agent):
        agent.mass *= self._mass_multiplier
        agent.c2 = 1. - (1. - agent.c2) * self._c2_multiplier

    def _revert_change(self, agent):
        agent.mass /= self._mass_multiplier
        agent.c2 = 1. - (1. - agent.c2) / self._c2_multiplier

    def step(self, state, meta_state):
        del meta_state

        self._steps_until_expire -= 1
        agent = state[self._agent_layer][0]

        if self._steps_until_expire == np.inf:
            boosters = state[self._booster_layer]
            if any([agent.overlaps_sprite(s) for s in boosters]):
                self._apply_change(agent)
                self._steps_until_expire = self.boost_duration
        elif self._steps_until_expire <= 0:
            self._revert_change(agent)
            self._steps_until_expire = np.inf


def get_config(_):
    """Get environment config."""

    ############################################################################
    # Sprite initialization
    ############################################################################

    # Agent
    agent_factors = distribs.Product(
        [distribs.Continuous('x', 0.1, 0.9),
         distribs.Continuous('y', 0.1, 0.9)],
        shape='circle', scale=0.1, c0=0.33, c1=1., c2=0.7,
    )

    # Predators
    predator_factors = distribs.Product(
        [distribs.Continuous('x', 0.1, 0.9),
         distribs.Continuous('y', 0.1, 0.9)],
        shape='circle', scale=0.1, c0=0., c1=1., c2=0.8,
    )

    # Prey
    prey_factors = distribs.Product(
        [distribs.Continuous('x', 0.1, 0.9),
         distribs.Continuous('y', 0.1, 0.9)],
        shape='circle', scale=0.1, c0=0.2, c1=1., c2=1.,
    )

    # Boosters
    booster_factors = distribs.Product(
        [distribs.Continuous('x', 0.1, 0.9),
         distribs.Continuous('y', 0.1, 0.9)],
        shape='triangle', scale=0.1, c0=0.6, c1=1., c2=1.,
    )

    # Portals
    portal_factors = dict(shape='square', scale=0.1, c0=0., c1=0., c2=0.95)
    portal_sprites = [
        sprite.Sprite(x=0.125, y=0.125, **portal_factors),
        sprite.Sprite(x=0.875, y=0.875, **portal_factors),
    ]

    # Walls
    wall_color = dict(c0=0., c1=0., c2=0.5)
    island_wall_shape_0 = np.array(
        [[0.2, 0.2], [0.4, 0.2], [0.4, 0.4], [0.2, 0.4]])
    island_wall_shapes = [
        island_wall_shape_0,
        island_wall_shape_0 + np.array([[0., 0.4]]),
        island_wall_shape_0 + np.array([[0.4, 0.4]]),
        island_wall_shape_0 + np.array([[0.4, 0.]]),
    ]
    island_walls = [
        sprite.Sprite(shape=shape, x=0., y=0., **wall_color)
        for shape in island_wall_shapes
    ]
    boundary_walls = shapes.border_walls(visible_thickness=0.05, **wall_color)
    walls = boundary_walls + island_walls

    # Callable sprite generators
    agent_generator = sprite_generators.generate_sprites(
        agent_factors, num_sprites=1)
    predator_generator = sprite_generators.generate_sprites(
        predator_factors, num_sprites=1)
    prey_generator = sprite_generators.generate_sprites(
        prey_factors, num_sprites=lambda: np.random.randint(2, 5))
    booster_generator = sprite_generators.generate_sprites(
        booster_factors, num_sprites=2)

    # Create callable initializer returning entire state
    def state_initializer():
        portals = portal_sprites
        agent = agent_generator(without_overlapping=walls)
        predators = predator_generator(without_overlapping=walls + agent)
        boosters = booster_generator(without_overlapping=walls + agent)
        prey = prey_generator(without_overlapping=walls)
        state = collections.OrderedDict([
            ('walls', walls),
            ('portals', portals),
            ('boosters', boosters),
            ('prey', prey),
            ('predators', predators),
            ('agent', agent),
        ])
        return state

    ############################################################################
    # Physics
    ############################################################################

    agent_friction_force = physics_lib.Drag(coeff_friction=0.25)
    predator_friction_force = physics_lib.Drag(coeff_friction=0.05)
    predator_random_force = physics_lib.RandomForce(max_force_magnitude=0.02)
    prey_friction_force = physics_lib.Drag(coeff_friction=0.02)
    prey_random_force = physics_lib.RandomForce(max_force_magnitude=0.02)
    predator_attraction = physics_lib.DistanceForce(
        force_fn=physics_lib.linear_force_fn(zero_intercept=-0.002, slope=0.001)
    )
    asymmetric_collision = physics_lib.Collision(
        elasticity=0.25, symmetric=False, update_angle_vel=False)
    
    forces = (
        (agent_friction_force, 'agent'),
        (predator_friction_force, 'predators'),
        (predator_random_force, 'predators'),
        (prey_friction_force, 'prey'),
        (prey_random_force, 'prey'),
        (predator_attraction, 'agent', 'predators'),
        (asymmetric_collision, ['agent', 'predators', 'prey'], 'walls'),
    )
    
    physics = physics_lib.Physics(*forces, updates_per_env_step=5)

    ############################################################################
    # Task
    ############################################################################

    predator_task = tasks.ContactReward(
        -5, layers_0='agent', layers_1='predators', reset_steps_after_contact=0)
    prey_task = tasks.ContactReward(1, layers_0='agent', layers_1='prey')
    reset_task = tasks.Reset(
        condition=lambda state: len(state['prey']) == 0,
        steps_after_condition=5,
    )
    task = tasks.CompositeTask(
        predator_task, prey_task, reset_task, timeout_steps=400)

    ############################################################################
    # Action space
    ############################################################################

    action_space = action_spaces.Joystick(
        scaling_factor=0.01,
        action_layers='agent',
    )

    ############################################################################
    # Observer
    ############################################################################

    observer = observers.PILRenderer(
        image_size=(64, 64),
        anti_aliasing=1,
        color_to_rgb='hsv_to_rgb',
    )

    ############################################################################
    # Game rules
    ############################################################################

    disappear_rule = game_rules.VanishOnContact(
        vanishing_layer='prey', contacting_layer='agent')
    portal_rule = game_rules.Portal(
        teleporting_layer='agent', portal_layer='portals')
    rules = (disappear_rule, portal_rule, Booster())

    ############################################################################
    # Final config
    ############################################################################

    config = {
        'state_initializer': state_initializer,
        'physics': physics,
        'task': task,
        'action_space': action_space,
        'observers': {'image': observer},
        'game_rules': rules,
    }
    return config



"""Maze class.

This file contains the Maze class, which is a class to mediate between a binary
maze array and sprites, including methods to infer a maze from sprites, convert
a maze to sprites, and sample points or connected blobs in the maze.
"""

import numpy as np
from moog import sprite

# Small error tolerance for checking whether wall vertex values are multiples of
# a grid size
_EPSILON = 1e-4

# Sanity check to prevent infinite while looping. No maze should have grid size
# larger than this
_MAX_MAZE_SIZE = 100


class Maze():
    """Maze class."""

    def __init__(self, maze):
        """Constructor.

        Args:
            maze: Square numpy array with binary values (may be a boolean array,
                or an int/float array with 0/1 values). The 1's are the walls of
                the maze.
        """
        self.maze = maze
        self.maze_size = maze.shape[0]
        self.grid_side = 1. / self.maze_size
        self.half_grid_side = 0.5 * self.grid_side
        self.side_vertices = np.linspace(
            self.half_grid_side, 1. - self.half_grid_side, self.maze_size)

    @classmethod
    def from_state(cls, state, maze_layer='walls'):
        """Get a maze object from environment state.

        This method works by first inferring the maze size by looking for the
        smallest integer N such that 1/N divides the vertex values of the
        sprites in the maze values. Then it checks whether each of the
        centerpoints of an NxN grid are contained in a maze wall sprite to
        compute the maze matrix.
        
        Args:
            state: OrderedDict of sprites. State of the environment.
            maze_layer: String. Layer in the environment containing the maze
                wall sprites.
        """
        wall_vertices = np.array([s.vertices for s in state[maze_layer]])
        
        # Find the smallest maze size N such that all wall vertices are a
        # multiple of 1/N
        maze_size = 1
        valid_maze_size = False
        while not valid_maze_size:
            if maze_size > _MAX_MAZE_SIZE:
                raise ValueError(
                    'Cannot find a maze grid size. Your maze sprites are '
                    'invalid.')
            rounded_vertices = np.round(wall_vertices * maze_size) / maze_size
            if np.allclose(rounded_vertices, wall_vertices, atol=_EPSILON):
                valid_maze_size = True
            else:
                maze_size += 1
        
        # Get a matrix of grid square centerpoints
        half_grid_side = 1. / (2 * maze_size)
        edge_centers = np.linspace(
            half_grid_side, 1 - half_grid_side, maze_size)
        grid_centers = np.stack(np.meshgrid(edge_centers, edge_centers), axis=2)
        flat_grid_centers = np.reshape(grid_centers, (maze_size * maze_size, 2))

        # Now check which grid square centerpoints are inside walls
        maze = np.zeros(maze_size * maze_size, dtype=bool)
        for s in state[maze_layer]:
            contained = s.contains_points(flat_grid_centers)
            maze = np.logical_or(maze, contained)

        maze = np.reshape(maze, (maze_size, maze_size)).astype(int)
        return cls(maze)

    def to_sprites(self, **color):
        """Turn a maze matrix into a list of sprites.
        
        Each sprite is a single square for one brick in the maze walls.

        Args:
            color: Dict. May contain keys {'c0', 'c1', 'c2'} for the wall sprite
                color factors.

        Returns:
            sprites: List of maze wall sprites.
        """
        maze_size = self.maze_size
        vertex_linspace = np.linspace(0., 1., maze_size + 1)
        sprites = []
        for x in range(maze_size):
            for y in range(maze_size):
                if self.maze[y, x]:
                    shape = np.array([
                        [vertex_linspace[x], vertex_linspace[y]],
                        [vertex_linspace[x], vertex_linspace[y + 1]],
                        [vertex_linspace[x + 1], vertex_linspace[y + 1]],
                        [vertex_linspace[x + 1], vertex_linspace[y]],
                    ])
                    new_sprite = sprite.Sprite(x=0., y=0., shape=shape, **color)
                    sprites.append(new_sprite)
        return sprites

    def open_vertex(self, i, j):
        """Returns whether the (i, j) cell is open, i.e. not a maze wall."""
        if i < 0 or j < 0 or i >= self.maze_size or j >= self.maze_size:
            return False
        else:
            return not self.maze[j, i]

    def valid_directions(self, i, j):
        """Computes the open neighbords of the (i, j) cell."""
        valid_directions = np.array([
            [self.open_vertex(k, j) for k in [i - 1, i + 1]],
            [self.open_vertex(i, k) for k in [j - 1, j + 1]],
        ])
        return valid_directions

    def sample_random_position(self, off_intersection=True):
        """Sample random open position on the edges of the maze."""
        # First find edges
        neg_maze = 1 - self.maze
        v_edges = np.logical_and(neg_maze[1:], neg_maze[:-1])
        h_edges = np.logical_and(neg_maze[:, 1:], neg_maze[:, :-1])
        v_edges = np.stack(np.nonzero(v_edges)[::-1]).T
        h_edges = np.stack(np.nonzero(h_edges)[::-1]).T
        num_h_edges = len(h_edges)
        num_v_edges = len(v_edges)
        num_edges = num_h_edges + num_v_edges
        if np.random.rand() < float(num_h_edges) / num_edges:
            # Pick a horizontal edge
            edge = h_edges[np.random.choice(num_h_edges)]
            position = edge
            if off_intersection:
                position = edge + np.random.rand() * np.array([1., 0.])
        else:
            # Pick a vertical edge
            edge = v_edges[np.random.choice(num_v_edges)]
            position = edge
            if off_intersection:
                position = edge + np.random.rand() * np.array([0., 1.])
        position = self.half_grid_side + position * self.grid_side
        return position

    def to_background_grid(self, line_thickness=0.01, **color):
        """Get static grid of background lines as list of sprites.

        These lines are in the channels of the maze and when rendered can
        provide a nice visual effect.
        
        Args:
            line_thickness: Float. How thick the grid lines should be.
            color: Dict. May contain keys {'c0', 'c1', 'c2'} for the background
                line sprite color factors.

        Returns:
            grid_sprites: List of sprites, the grid lines.
        """
        grid_sprites = []
        def _add_sprite(min_x, max_x, min_y, max_y):
            shape = np.array([
                [min_x, min_y], [max_x, min_y], [max_x, max_y], [min_x, max_y]
            ])
            grid_sprites.append(sprite.Sprite(x=0., y=0., shape=shape, **color))
        
        for x in self.side_vertices:
            min_x = x - 0.5 * line_thickness
            max_x = x + 0.5 * line_thickness
            _add_sprite(min_x, max_x, 0., 1.)

        for y in self.side_vertices:
            min_y = y - 0.5 * line_thickness
            max_y = y + 0.5 * line_thickness
            _add_sprite(0., 1., min_y, max_y)

        return grid_sprites

    def sample_open_point(self):
        """Sample an open point in the maze.

        Returns:
            Tuple of integers (i, j), coordinates of open point.
        """
        if np.sum(1 - self.maze) == 0:
            raise ValueError('Maze has no open point.')
        candidates = np.argwhere(self.maze == 0)
        point = candidates[np.random.randint(len(candidates))]
        return tuple(point)

    def sample_distinct_open_points(self, num_points):
        """Sample distinct open points in the maze.

        Args:
            num_points: Int. Number of distinct open points to sample.

        Returns:
            Tuple of integers (i, j), coordinates of open point.
        """
        if np.sum(1 - self.maze) < num_points:
            raise ValueError('Maze has no open point.')
        candidates = np.argwhere(self.maze == 0)
        inds = np.random.choice(len(candidates), size=num_points, replace=False)
        points = [tuple(candidates[i]) for i in inds]
        return points
    
    def get_neighbors(self, i, j):
        """Get list of neighbors of point (i, j)."""
        neighbors = []
        if i > 0 and not self.maze[i - 1, j]:
            neighbors.append((i - 1, j))
        if i < self.maze_size - 1 and not self.maze[i + 1, j]:
            neighbors.append((i + 1, j))
        if j > 0 and not self.maze[i, j - 1]:
            neighbors.append((i, j - 1))
        if j < self.maze_size - 1 and not self.maze[i, j + 1]:
            neighbors.append((i, j + 1))
        return neighbors
    
    def get_neighbor_dict(self):
        """Get dictionary of neighbors of all points.
        
        Returns:
            neighbor_dict: Keys are int tuples (i, j) and values are lists of
                int tuples for all open neighbors of (i, j).
        """
        neighbor_dict = {
            (i, j): self.get_neighbors(i, j)
            for i in range(self.maze_size)
            for j in range(self.maze_size)
        }
        return neighbor_dict

    def add_wall(self, x_range, y_range):
        """Add a rectangular wall to the maze.

        Args:
            x_range: start and end point of the wall in the x-coordinate
            y_range: start and end point of the wall in the y-coordinate
        """
        self.maze[x_range[0] : x_range[1] + 1, y_range[0] : y_range[1] + 1] = 1

    def add_outer_walls(self):
        """Make the maze borders walls.
        
        Warning: This could distrupt properties of the maze. For example, if the
        maze originally had no dead ends, this could introduce dead ends.
        """
        self.maze[0, :] = 1
        self.maze[-1, :] = 1
        self.maze[:, 0] = 1
        self.maze[:, -1] = 1



"""This file contains GUI frames for human_agent.py.

The classes in this file are interfaces for playing the demo with different
action spaces. If you are using an action space that doesn't fall into one of
these categories, you must create your own gui for it.

Note: on some computers, the keyboard interfaces don't work properly, and
holding letter or arrow keys repeatedly takes actions. This is an issue with
your computer's keyboard sending rapid press/release signals when you hold down
a key. You can resolve this by (i) changing your computer's keyboard settings to
not do this behavior when you hold down a key, or (ii) modify the class in this
file that you're using to do whatever behavior you want when a key is held down
(this will involve a bit of debugging for you).
"""

import numpy as np
import tkinter as tk


class JoystickFrame(tk.Frame):
    """Joystick Tkinter frame.

    This creates the frame for an interactive joystick. The joystick consists of
    three objects:
        (i) A large gray "motion zone" circle in the background. This is the
            area in which the joystick can be moved.
        (ii) A small black "center point" circle fixed in the middle ground.
            This indicates the joystick position that give zero action.
        (iii) A green "joystick" circle in the foreground that can be moved. The
            center of this circle is the action readout.

    If the mouse is not clicked, then position of the joystick centerpoint is 0
    (in the center of the motion zone). If the mouse is currently pressed, then
    the position of the joystick is the closest point in the motion zone to the
    mouse position. Namely, if the mouse is in the motion zone then the joystick
    is directly under the mouse, and if the mouse is not in the motion zone then
    the joystick is at the edge of the motion zone closest to the mouse.

    Thus the joystick can be moved by clicking the mouse anywhere (in which case
    the joystick jumps to that position) and dragging the pressed mouse around
    (in which case the joystick moves underneath the mouse).
    """

    def __init__(self, root, canvas_half_width=100, motion_zone_radius=90,
                 joystick_radius=10, center_point_radius=3):
        """Constructor.

        Args:
            root: Instance of tk.Frame. Root frame in which the joystick lives.
            canvas_half_width: Int. Half of the width of the canvas on which the
                joystick is rendered.
            motion_zone_radius: Int. Radius of the motion zone.
            joystick_radius: Int. Radius of the joystick.
            center_point_radius: Int. Radius of the center point.
        """
        super(JoystickFrame, self).__init__(root)

        self._joystick_radius = joystick_radius
        self._canvas_half_width = canvas_half_width
        self._motion_zone_radius = motion_zone_radius
        self._center_point_radius = center_point_radius

        # Create a canvas
        self.canvas = tk.Canvas(
            width=2 * canvas_half_width,
            height=2 * canvas_half_width)

        # Create motion zone, center point, and joystick
        self._create_items()

        # Add bindings for clicking, dragging and releasing the joystick
        self.canvas.bind('<ButtonPress-1>', self._mouse_press)
        self.canvas.bind('<ButtonRelease-1>', self._mouse_release)
        self.canvas.bind('<B1-Motion>', self._mouse_move)

        self._mouse_is_pressed = False

    def _create_items(self):
        # Create motion zone
        self.canvas.create_oval(
            self._canvas_half_width - self._motion_zone_radius,
            self._canvas_half_width - self._motion_zone_radius,
            self._canvas_half_width + self._motion_zone_radius,
            self._canvas_half_width + self._motion_zone_radius,
            fill='gray80',
        )

        # Create center point
        self.canvas.create_oval(
            self._canvas_half_width - self._center_point_radius,
            self._canvas_half_width - self._center_point_radius,
            self._canvas_half_width + self._center_point_radius,
            self._canvas_half_width + self._center_point_radius,
            fill='black',
        )

        # Create joystick
        self.joystick = self.canvas.create_oval(
            self._canvas_half_width - self._joystick_radius,
            self._canvas_half_width - self._joystick_radius,
            self._canvas_half_width + self._joystick_radius,
            self._canvas_half_width + self._joystick_radius,
            fill='green',
        )

    def _recenter_joystick(self):
        """Move the joystick to the center."""
        new_coords = [
            self._canvas_half_width - self._joystick_radius,
            self._canvas_half_width - self._joystick_radius,
            self._canvas_half_width + self._joystick_radius,
            self._canvas_half_width + self._joystick_radius,
        ]
        self.canvas.coords(self.joystick, new_coords)

    def _place_joystick(self, event):
        """Place the joystick near the (x, y) coordinates of a mouse event.

        If the event (x, y) is inside the motion zone, the joystick is placed
        directly at that position. If it is outside the motion zone, the
        joystick is placed on the edge of the motion zone nearest to that point.
        """
        centered_event_coords = (
            np.array([event.x, event.y], dtype=float) - self._canvas_half_width)
        event_dist = np.linalg.norm(centered_event_coords)
        rescale_factor = min(1, self._motion_zone_radius / event_dist)
        centered_event_coords *= rescale_factor
        event_coords = centered_event_coords + self._canvas_half_width
        event_coords = event_coords.astype(int)

        new_coords = [
            event_coords[0] - self._joystick_radius,
            event_coords[1] - self._joystick_radius,
            event_coords[0] + self._joystick_radius,
            event_coords[1] + self._joystick_radius,
        ]
        self.canvas.coords(self.joystick, new_coords)

    def _mouse_press(self, event):
        self._place_joystick(event)
        self._mouse_is_pressed = True

    def _mouse_release(self, event):
        self._mouse_is_pressed = False
        self._recenter_joystick()

    def _mouse_move(self, event):
        if self._mouse_is_pressed:
            self._place_joystick(event)

    @property
    def action(self):
        """Return the joystick's position as an action in [-1, 1] x [-1, 1]."""
        joystick_coords = self.canvas.coords(self.joystick)
        joystick_center = np.array([
            joystick_coords[0] + self._joystick_radius,
            joystick_coords[1] + self._joystick_radius
        ])
        joystick_center -= self._canvas_half_width
        action = joystick_center.astype(float) / self._motion_zone_radius
        return np.array([action[0], -1 * action[1]])


class GridActions(tk.Frame):
    """Grid actions Tkinter frame.

    This creates an empty Tkinter frame where the joystick would be. It also
    registers bindings responding to arrow key presses and releases, and turns
    them into discrete actions for a Grid action space.
    """

    def __init__(self, root, canvas_half_width=100):
        """Constructor.

        Args:
            root: Instance of tk.Frame. Root frame in which the gui frame lives.
            canvas_half_width: Int. Half of the width of the canvas to create.
        """
        super(GridActions, self).__init__(root)
        self._current_key = 4  # Do-nothing action

        # Create a canvas
        self.canvas = tk.Canvas(
            width=2 * canvas_half_width,
            height=2 * canvas_half_width)

        # Add bindings for key presses and releases
        root.bind('<KeyPress>', self._key_press)
        root.bind('<KeyRelease>', self._key_release)

    def _get_action_from_event(self, event):
        if event.keysym == 'Left':
            return 0
        elif event.keysym == 'Right':
            return 1
        elif event.keysym == 'Down':
            return 2
        elif event.keysym == 'Up':
            return 3
        else:
            return None

    def _key_press(self, event):
        self._current_key = self._get_action_from_event(event)

    def _key_release(self, event):
        if self._get_action_from_event(event) == self._current_key:
            self._current_key = None

    @property
    def action(self):
        if self._current_key is not None:
            return self._current_key
        else:
            return 4  # Do-nothing action


class SetPositionFrame():
    """SetPosition Tkinter frame.

    This creates position_setting functionality for a canvas. Typically this is
    done to the HumanAgent._env_canvas.
    """
    def __init__(self, canvas, canvas_half_width):
        """Constructor.

        Args:
            canvas: Canvas object to add position-setting functionality to.
            canvas_half_width: Int. Half-width of the canvas.
        """
        # Add bindings for clicking, dragging and releasing the joystick
        canvas.bind('<ButtonPress-1>', self._mouse_press)
        canvas.bind('<ButtonRelease-1>', self._mouse_release)
        canvas.bind('<B1-Motion>', self._mouse_move)

        self._canvas_half_width = canvas_half_width
        self._mouse_is_pressed = False
        self._mouse_coords = np.array([0.5, 0.5])

    def _mouse_press(self, event):
        self._place_mouse(event)
        self._mouse_is_pressed = True

    def _mouse_release(self, event):
        self._mouse_is_pressed = False

    def _mouse_move(self, event):
        if self._mouse_is_pressed:
            self._place_mouse(event)

    def _place_mouse(self, event):
        """Place the self._mouse_coords (x, y) coordinates of a mouse event."""
        centered_event_coords = (
            np.array([event.x, event.y], dtype=float) - self._canvas_half_width)
        centered_event_coords = np.clip(
            centered_event_coords,
            -self._canvas_half_width,
            self._canvas_half_width,
        )
        self._mouse_coords = 0.5 * (
            1 + centered_event_coords.astype(float) / self._canvas_half_width)

    @property
    def action(self):
        """Return the mouse's position as an action in [0, 1] x [0, 1]."""
        return np.array([self._mouse_coords[0], 1. - self._mouse_coords[1]])


class TwoPlayerGridActions(tk.Frame):
    """2-player grid actions Tkinter frame.

    This creates an empty Tkinter frame where the joystick would be. It also
    registers bindings responding to arrow key presses and releases, and turns
    them into discrete actions for a two-player Grid action space.

    One player uses the [Left, Right, Down, Up] arrow keys and the other player
    uses the [a, d, s, w] keys.
    """

    def __init__(self, root, canvas_half_width=100, player_0='', player_1=''):
        """Constructor.

        Args:
            root: Instance of tk.Frame. Root frame in which the gui frame lives.
            canvas_half_width: Int. Half of the width of the canvas to create.
            player_0: String. Name of player_0. This will be the key in the
                action dictionary for player_0's action.
            player_1: String. Name of player_1. This will be the key in the
                action dictionary for player_1's action.
        """
        super(TwoPlayerGridActions, self).__init__(root)
        self._player_0 = player_0
        self._player_1 = player_1
        self._current_keys = {player_0: 4, player_1: 4}  # Do-nothing action

        # Create a canvas
        self.canvas = tk.Canvas(
            width=2 * canvas_half_width,
            height=2 * canvas_half_width)

        # Add bindings for key presses and releases
        root.bind('<KeyPress>', self._key_press)
        root.bind('<KeyRelease>', self._key_release)

    def _get_action_from_event(self, event):
        if event.keysym == 'Left':
            return 0
        elif event.keysym == 'Right':
            return 1
        elif event.keysym == 'Down':
            return 2
        elif event.keysym == 'Up':
            return 3
        elif event.keysym == 'a':
            return 4
        elif event.keysym == 'd':
            return 5
        elif event.keysym == 's':
            return 6
        elif event.keysym == 'w':
            return 7
        else:
            return None

    def _key_press(self, event):
        current_key = self._get_action_from_event(event)
        if current_key is None:
            pass
        elif current_key < 4:
            self._current_keys[self._player_0] = current_key
        else:
            self._current_keys[self._player_1] = current_key - 4

    def _key_release(self, event):
        if (self._get_action_from_event(event) ==
                self._current_keys[self._player_0]):
            self._current_keys[self._player_0] = 4
        elif (self._get_action_from_event(event) ==
                self._current_keys[self._player_1] + 4):
            self._current_keys[self._player_1] = 4

    @property
    def action(self):
        return self._current_keys


class TwoPlayerJoystick(tk.Frame):
    """2-player joystick actions Tkinter frame.

    This creates an empty Tkinter frame where the joystick would be. It also
    registers bindings responding to arrow key presses and releases, and turns
    them into joystick actions for a two-player Joystick action space.

    One player uses the [Left, Right, Down, Up] arrow keys and the other player
    uses the [a, d, s, w] keys.
    """

    def __init__(self, root, canvas_half_width=100, player_0='', player_1=''):
        """Constructor.

        Args:
            root: Instance of tk.Frame. Root frame in which the gui frame lives.
            canvas_half_width: Int. Half of the width of the canvas to create.
            player_0: String. Name of player_0. This will be the key in the
                action dictionary for player_0's action.
            player_1: String. Name of player_1. This will be the key in the
                action dictionary for player_1's action.
        """
        super(TwoPlayerJoystick, self).__init__(root)
        self._player_0 = player_0
        self._player_1 = player_1
        self._current_action = {
            player_0: np.array([0.,0.]),
            player_1: np.array([0.,0.]),
        }
        self._current_keys = {
            player_0: [0.,0.,0.,0.],
            player_1: [0.,0.,0.,0.],
        }

        # Create a canvas
        self.canvas = tk.Canvas(
            width=2 * canvas_half_width,
            height=2 * canvas_half_width)

        # Add bindings for key presses and releases
        root.bind('<KeyPress>', self._key_press)
        root.bind('<KeyRelease>', self._key_release)

    def _key_press(self, event):
        if event.keysym == 'Left':
            self._current_keys[self._player_0][0] = 1.
        elif event.keysym == 'Right':
            self._current_keys[self._player_0][1] = 1.
        elif event.keysym == 'Down':
            self._current_keys[self._player_0][2] = 1.
        elif event.keysym == 'Up':
            self._current_keys[self._player_0][3] = 1.
        elif event.keysym == 'a' or event.keysym == 'Meta_L':
            self._current_keys[self._player_1][0] = 1.
        elif event.keysym == 'd' or event.keysym == 'Alt_L':
            self._current_keys[self._player_1][1] = 1.
        elif event.keysym == 's':
            self._current_keys[self._player_1][2] = 1.
        elif event.keysym == 'w':
            self._current_keys[self._player_1][3] = 1.
        for k, v in self._current_keys.items():
            self._current_action[k] = np.array([v[1]-v[0],v[3]-v[2]])

    def _key_release(self, event):
        if event.keysym == 'Left':
            self._current_keys[self._player_0][0] = 0.
        elif event.keysym == 'Right':
            self._current_keys[self._player_0][1] = 0.
        elif event.keysym == 'Down':
            self._current_keys[self._player_0][2] = 0.
        elif event.keysym == 'Up':
            self._current_keys[self._player_0][3] = 0.
        elif event.keysym == 'a' or event.keysym == 'Meta_L':
            self._current_keys[self._player_1][0] = 0.
        elif event.keysym == 'd' or event.keysym == 'Alt_L':
            self._current_keys[self._player_1][1] = 0.
        elif event.keysym == 's':
            self._current_keys[self._player_1][2] = 0.
        elif event.keysym == 'w':
            self._current_keys[self._player_1][3] = 0.
        for k, v in self._current_keys.items():
            self._current_action[k] = np.array([v[1]-v[0],v[3]-v[2]])

    @property
    def action(self):
        return self._current_action



# Copyright 2019 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Functions to transform between color spaces for rendering."""

import colorsys
import numpy as np


def hsv_to_rgb(c):
  """Convert HSV tuple to RGB tuple."""
  return tuple((255 * np.array(colorsys.hsv_to_rgb(*c))).astype(np.uint8))



"""Multi-object tracking with features.

In this task, the subject controls the position of a small cross. Typically,
this would be controlled by an eye-tracker to follow the subject's gaze. There
are a number of circles with oriented bars (each initially randomly vertical or
horizontal). Those circles with bars drift around an arena, bouncing off the
walls. At a random time, one of the bars rotates 90 degrees. The subject's goal
is to identify and fixate on that target.

After a brief initial period in which they are all visible, an occluder
appears occluding peripheral vision. This forces the subject to mentally keep
track of the targets' locations and orientations.
"""

import collections
import numpy as np

from moog import action_spaces
from moog import game_rules as gr
from moog import observers
from moog import physics as physics_lib
from moog import shapes
from moog import sprite
from moog import tasks
from moog.state_initialization import distributions as distribs

_FIXATION_THRESHOLD = 0.1


class RadialVelocity(distribs.AbstractDistribution):
    """Radial velocity distribution."""

    def __init__(self, speed):
        """Constructor.

        Args:
            speed: Float. Speed of the sampled velocities.
        """
        self._speed = speed

    def sample(self, rng):
        rng = self._get_rng(rng)
        theta = rng.uniform(0., 2 * np.pi)
        x_vel = self._speed * np.cos(theta)
        y_vel = self._speed * np.sin(theta)
        return {'x_vel': x_vel, 'y_vel': y_vel}

    def contains(self, spec):
        if 'x_vel' not in spec or 'y_vel' not in spec:
            return False

        vel_norm = np.linalg.norm([spec['x_vel'], spec['y_vel']])
        if np.abs(vel_norm - self._speed) < _EPSILON:
            return True
        else:
            return False

    def to_str(self, indent):
        s = 'RadialVelocity({})'.format(self._speed)
        return indent * '  ' + s

    @property
    def keys(self):
        return set(['x_vel', 'y_vel'])


class ChangeTargetFeature(gr.AbstractRule):
    """Game rule to change the first target's oriented bar."""

    def step(self, state, meta_state):
        del meta_state
        """Rotate first target's bar np 90 degrees."""
        s = state['bars'][0]
        s.angle = s.angle + 0.5 * np.pi


def get_config(num_targets):
    """Get environment config.
    
    Args:
        num_targets: Int. Number of targets.
    """

    ############################################################################
    # Sprite initialization
    ############################################################################

    # Target circles
    target_factors = distribs.Product(
        [distribs.Continuous('x', 0.1, 0.9),
         distribs.Continuous('y', 0.1, 0.9),
         RadialVelocity(speed=0.01),],
        scale=0.1, shape='circle', c0=0., c1=0., c2=0.9,
    )

    # Target bars
    bar_factors = dict(
        scale=0.1, shape='square', aspect_ratio=0.3, c0=0., c1=0., c2=0.2)

    # Walls
    bottom_wall = [[-1, 0], [2, 0], [2, -1], [-1, -1]]
    top_wall = [[-1, 1], [2, 1], [2, 2], [-1, 2]]
    left_wall = [[0, -1], [0, 4], [-1, 4], [-1, -1]]
    right_wall = [[1, -1], [1, 4], [2, 4], [2, -1]]
    walls = [
        sprite.Sprite(shape=np.array(v), x=0, y=0, c0=0., c1=0., c2=0.5)
        for v in [bottom_wall, top_wall, left_wall, right_wall]
    ]

    # Occluder
    occluder_factors = dict(x=0.5, y=0.5, c0=0.6, c1=0.25, c2=0.5, opacity=0)

    # Cross shape for agent and fixation cross
    cross_shape = 0.1 * np.array([
        [-5, 1], [-1, 1], [-1, 5], [1, 5], [1, 1], [5, 1], [5, -1], [1, -1],
        [1, -5], [-1, -5], [-1, -1], [-5, -1]
    ])
    
    def state_initializer():

        fixation = sprite.Sprite(
            x=0.5, y=0.5, shape=cross_shape, scale=0.1, c0=0., c1=0., c2=0.)
        screen = sprite.Sprite(
            x=0.5, y=0.5, shape='square', scale=2., c0=0., c1=0., c2=1.)
            
        agent = sprite.Sprite(
            x=0.5, y=0.5, scale=0.04, shape=cross_shape, c0=0.33, c1=1., c2=1.)
        occluder_shape = shapes.annulus_vertices(0.13, 2.)
        occluder = sprite.Sprite(shape=occluder_shape, **occluder_factors)

        targets = [
            sprite.Sprite(**target_factors.sample())
            for _ in range(num_targets)
        ]

        bar_angles = 0.5 * np.pi * np.random.binomial(1, 0.5, (num_targets))
        bars = [
            sprite.Sprite(
                x=s.x, y=s.y, x_vel=s.x_vel, y_vel=s.y_vel, angle=angle,
                **bar_factors)
            for s, angle in zip(targets, bar_angles)
        ]

        state = collections.OrderedDict([
            ('walls', walls),
            ('targets', targets),
            ('bars', bars),
            ('occluder', [occluder]),
            ('screen', [screen]),
            ('fixation', [fixation]),
            ('agent', [agent]),
        ])
        return state

    ############################################################################
    # Physics
    ############################################################################

    elastic_collision = physics_lib.Collision(
        elasticity=1., symmetric=False, update_angle_vel=False)
    tether = physics_lib.TetherZippedLayers(
        layer_names=('targets', 'bars'), update_angle_vel=False)
    physics = physics_lib.Physics(
        (elastic_collision, 'targets', 'walls'),
        updates_per_env_step=10,
        corrective_physics=[tether],
    )

    ############################################################################
    # Task
    ############################################################################

    def _reward_condition(_, meta_state):
        return meta_state['phase'] == 'reward'
    task = tasks.Reset(
        condition=_reward_condition, reward_fn=lambda _: 1,
        steps_after_condition=10,
    )

    ############################################################################
    # Action space
    ############################################################################

    action_space = action_spaces.SetPosition(
        action_layers=('agent', 'occluder'))

    ############################################################################
    # Observer
    ############################################################################

    observer = observers.PILRenderer(
        image_size=(64, 64),
        anti_aliasing=1,
        color_to_rgb=observers.color_maps.hsv_to_rgb,
    )

    ############################################################################
    # Game rules
    ############################################################################

    # Fixation phase

    fixation_rule = gr.Fixation(
        'agent', 'fixation', _FIXATION_THRESHOLD, 'fixation_duration')
    def _should_end_fixation(_, meta_state):
        return meta_state['fixation_duration'] >= 15

    fixation_phase = gr.Phase(
        continual_rules=fixation_rule,
        end_condition=_should_end_fixation,
        name='fixation',
    )

    # Visible Phase
    
    vanish_fixation = gr.VanishByFilter('fixation', lambda _: True)
    vanish_screen = gr.VanishByFilter('screen', lambda _: True)

    visible_phase = gr.Phase(
        one_time_rules=[vanish_fixation, vanish_screen],
        duration=5,
        name='visible',
    )

    # Tracking Phase

    def _make_opaque(s):
        s.opacity = 255
    appear_occluder = gr.ModifySprites('occluder', _make_opaque)

    tracking_phase = gr.Phase(
        one_time_rules=appear_occluder,
        duration=lambda: np.random.randint(40, 80),
        name='tracking',
    )

    # Change Phase
    
    fixation_response_rule = gr.Fixation(
        'agent', 'targets', _FIXATION_THRESHOLD, 'response_duration')
    def _should_end_change(_, meta_state):
        return meta_state['response_duration'] >= 30

    change_phase = gr.Phase(
        one_time_rules=ChangeTargetFeature(),
        continual_rules=fixation_response_rule,
        name='change',
        end_condition=_should_end_change,
    )

    # Reward Phase

    def _make_transparent(s):
        s.opacity = 0
    disappear_occluder = gr.ModifySprites('occluder', _make_transparent)
    def _glue(s):
        s.velocity = np.zeros(2)
    glue_targets = gr.ModifySprites(('targets', 'bars'), _glue)
    
    reward_phase = gr.Phase(
        one_time_rules=(disappear_occluder, glue_targets),
        name='reward',
    )

    phase_sequence = gr.PhaseSequence(
        fixation_phase, visible_phase, tracking_phase, change_phase,
        reward_phase, meta_state_phase_name_key='phase',
    )

    ############################################################################
    # Final config
    ############################################################################

    config = {
        'state_initializer': state_initializer,
        'physics': physics,
        'task': task,
        'action_space': action_space,
        'observers': {'image': observer, 'state': observers.RawState()},
        'game_rules': (phase_sequence,),
        'meta_state_initializer': lambda: {'phase': ''},
    }
    return config
